﻿using System;

namespace GenericMethodsAndInheritance {

	class X {
		public void m(decimal t) { Console.WriteLine("X.m(decimal)"); }
	}

	class A : X {
		public void m(bool t) { Console.WriteLine("A.m(bool)"); }
		public virtual void m<T>(T t) { Console.WriteLine("A.m: T == {0}", typeof(T)); }
	}

	class B : A {
		public override void m<T>(T t) { Console.WriteLine("B.m: T == {0}", typeof(T)); }
	}

	class C : A {
		// Warning: The member 'C.m(double)' does not hide an inherited member. The new keyword is not required.
		public new void m(double t) { Console.WriteLine("C.m(double)"); }
	}

	class D : B {
		// Warning: The member 'D.m(int)' does not hide an inherited member. The new keyword is not required.
		public new void m(int t) { Console.WriteLine("D.m(int)"); }
	}

	class Program {
		static void Main(string[] args) {
			A a = new A();
			B b = new B();
			C c = new C();
			D d = new D();

			Console.WriteLine("+++ a.m (A): ");
			a.m(1);
			a.m(1.0);

			Console.WriteLine("+++ b.m (B): ");
			b.m(1);
			b.m(1.0);

			Console.WriteLine();

			// KEY CODE STARTS HERE:
			Console.WriteLine("+++ c.m (C): ");
			c.m(1);			// Use single arg method from C -> have one = m(double) -> use it with implicit conversion from int
			c.m(1.0);		// Use single arg method from C -> have one = m(double) -> use it
			c.m("Hello");   // Use single arg method from C -> have one = m(double) -> NOT COMPATIBLE WITH STRING -> try "2 argument" method -> have one m<T>(T) [1 type arg + 1 normal arg] -> use it
			c.m(true);
			c.m(1M);
			c.m('a');

			Console.WriteLine("+++ d.m (D): ");
			d.m(1);		// Use single arg method from D -> have one = m(int) -> use it
			d.m(1.0);   // Use single arg method from D -> have one = m(int) -> NOT COMPATIBLE WITH DOUBLE -> try "2 argument" method -> have one m<T>(T) [1 type arg + 1 normal arg] -> use it

			Console.WriteLine();

			a = b;
			Console.WriteLine("+++ a.m (B): ");
			a.m(1);
			a.m(1.0);
			a.m(1M);
			a.m('a');

			a = c;
			Console.WriteLine("+++ a.m (C): ");
			a.m(1);
			a.m(1.0);
			a.m(1M);
			a.m('a');

			a = d;
			Console.WriteLine("+++ a.m (D): ");
			a.m(1);
			a.m(1.0);
			a.m(1M);
			a.m('a');

			X x = c;
			Console.WriteLine("+++ x.m (C): ");
			x.m(1M);

		}
	}
}
