﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;

namespace ExtensionMethodsGettingGeneric {

	static class IComparableExtensions {
		public static bool IsSameAs<T>(this T first, T second) where T : IComparable<T>
			=> first.CompareTo(second) == 0;
	}

	class A : IComparable<A> {
		public int X;
		public int CompareTo(A other) => this.X.CompareTo(other.X);
	}

	class B {
		public int X;
		public int CompareTo(A other) => this.X.CompareTo(other.X);
	}

	class Program {
		static void Main(string[] args) {
			Console.WriteLine(1.IsSameAs(2));
			Console.WriteLine("Hello".IsSameAs("He" + "llo"));
            Console.WriteLine();
            Console.WriteLine(new A { X = 10 }.IsSameAs(new A { X = 5 + 5 }));

			// Error: The type 'ExtensionMethodsGettingGeneric.B' cannot be used as type parameter 'T' 
			//        in the generic type or method 'IComparableExtensions.IsSameAs<T>(T, T)'.
			//        There is no implicit reference conversion from 'ExtensionMethodsGettingGeneric.B'
			//        to 'System.IComparable<ExtensionMethodsGettingGeneric.B>'.

			// Console.WriteLine(new B { X = 10 }.IsSameAs(new B { X = 5 + 5 }));

			Console.WriteLine(new C { X = 10 }.IsSameAs(new C { X = 5 + 5 }));
		}
	}

	class C : IComparable<C> {
		public int X;
		public int CompareTo(C other) {
			Console.WriteLine("C.CompareTo()");
			return this.X.CompareTo(other.X);
		}

		public bool IsSameAs(C other) {
			Console.WriteLine("C.IsSameAs()");
			return this.X == other.X;
		}
	}
}
